from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.tasks.salesforce import BaseRetrieveMetadata


retrieve_unpackaged_options = BaseRetrieveMetadata.task_options.copy()
retrieve_unpackaged_options.update({
    'package_xml': {
        'description': 'The path to a package.xml manifest to use for the retrieve.',
        'required': True,
    },
    'api_version': {
        'description': (
            'Override the default api version for the retrieve.' +
            ' Defaults to project__package__api_version'
        ),
    },
})

class RetrieveUnpackaged(BaseRetrieveMetadata):
    api_class = ApiRetrieveUnpackaged

    task_options = retrieve_unpackaged_options

    def _init_options(self, kwargs):
        super(RetrieveUnpackaged, self)._init_options(kwargs)

        if 'package_xml' in self.options:
            self.options['package_xml_path'] = self.options['package_xml']
            with open(self.options['package_xml_path'], 'r') as f:
                self.options['package_xml'] = f.read()

    def _get_api(self):
        return self.api_class(
            self,
            self.options['package_xml'],
            self.options.get('api_version'),
        )



import os
import shutil
import tempfile

from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.tasks.salesforce import Deploy
from cumulusci.utils import CUMULUSCI_PATH
from cumulusci.utils import findReplace


class UpdateAdminProfile(Deploy):
    name = 'UpdateAdminProfile'

    task_options = {
        'package_xml': {
            'description': 'Override the default package.xml file for retrieving the Admin.profile and all objects and classes that need to be included by providing a path to your custom package.xml',
        }
    }

    def _init_options(self, kwargs):
        super(UpdateAdminProfile, self)._init_options(kwargs)

        if 'package_xml' not in self.options:
            self.options['package_xml'] = os.path.join(CUMULUSCI_PATH, 'cumulusci', 'files', 'admin_profile.xml')

        self.options['package_xml_path'] = self.options['package_xml']
        with open(self.options['package_xml_path'], 'r') as f:
            self.options['package_xml'] = f.read()

    def _run_task(self):
        self.tempdir = tempfile.mkdtemp()
        self._retrieve_unpackaged()
        self._process_metadata()
        self._deploy_metadata()
        shutil.rmtree(self.tempdir)

    def _retrieve_unpackaged(self):
        self.logger.info('Retrieving metadata using {}'.format(self.options['package_xml_path']))
        api_retrieve = ApiRetrieveUnpackaged(
            self,
            self.options.get('package_xml'),
            self.project_config.project__package__api_version,
        )
        unpackaged = api_retrieve()
        unpackaged.extractall(self.tempdir)

    def _process_metadata(self):
        self.logger.info('Processing retrieved metadata in {}'.format(self.tempdir))

        findReplace(
            '<editable>false</editable>',
            '<editable>true</editable>',
            os.path.join(self.tempdir, 'profiles'),
            'Admin.profile',
        )
        findReplace(
            '<readable>false</readable>',
            '<readable>true</readable>',
            os.path.join(self.tempdir, 'profiles'),
            'Admin.profile',
        )

    def _deploy_metadata(self):
        self.logger.info('Deploying updated Admin.profile from {}'.format(self.tempdir))
        api = self._get_api(path=self.tempdir)
        return api()



from cumulusci.salesforce_api.package_zip import UninstallPackageZipBuilder
from cumulusci.tasks.salesforce import Deploy


class UninstallPackage(Deploy):
    task_options = {
        'namespace': {
            'description': 'The namespace of the package to uninstall.  Defaults to project__package__namespace',
            'required': True,
        },
        'purge_on_delete': {
            'description': 'Sets the purgeOnDelete option for the deployment.  Defaults to True',
            'required': True,
        },
    }

    def _init_options(self, kwargs):
        super(UninstallPackage, self)._init_options(kwargs)
        if 'namespace' not in self.options:
            self.options['namespace'] = self.project_config.project__package__namespace
        if 'purge_on_delete' not in self.options:
            self.options['purge_on_delete'] = True
        if self.options['purge_on_delete'] == 'False':
            self.options['purge_on_delete'] = False

    def _get_api(self, path=None):
        package_zip = UninstallPackageZipBuilder(
            self.options['namespace'],
            self.project_config.project__package__api_version,
        )
        return self.api_class(self, package_zip(), purge_on_delete=self.options['purge_on_delete'])



import os
import sys
import click

from cumulusci.core.config import YamlGlobalConfig
from cumulusci.core.config import YamlProjectConfig
from cumulusci.core.exceptions import ConfigError
from cumulusci.core.exceptions import NotInProject
from cumulusci.core.exceptions import ProjectConfigNotFound
from cumulusci.core.utils import import_class
from .logger import init_logger


class CliConfig(object):

    def __init__(self):
        self.global_config = None
        self.project_config = None
        self.keychain = None

        init_logger()
        self._load_global_config()
        self._load_project_config()
        self._load_keychain()
        self._add_repo_to_path()

    def _add_repo_to_path(self):
        if self.project_config:
            sys.path.append(self.project_config.repo_root)

    def _load_global_config(self):
        try:
            self.global_config = YamlGlobalConfig()
        except NotInProject as e:
            raise click.UsageError(e.message)

    def _load_project_config(self):
        try:
            self.project_config = self.global_config.get_project_config()
        except ProjectConfigNotFound:
            pass
        except NotInProject as e:
            raise click.UsageError(e.message)
        except ConfigError as e:
            raise click.UsageError('Config Error: {}'.format(e.message))

    def _load_keychain(self):
        self.keychain_key = os.environ.get('CUMULUSCI_KEY')
        if self.project_config:
            keychain_class = os.environ.get(
                'CUMULUSCI_KEYCHAIN_CLASS',
                self.project_config.cumulusci__keychain,
            )
            self.keychain_class = import_class(keychain_class)
            self.keychain = self.keychain_class(
                self.project_config, self.keychain_key)
            self.project_config.set_keychain(self.keychain)




import base64
import hashlib
import io
import logging
import os

from cumulusci.core.exceptions import GithubException


class CommitDir(object):
    """Commit all changes in local_dir to branch/repo_dir"""

    def __init__(self, repo, logger=None, author=None):
        """
        repo: github3.py Repository object, authenticated
        logger (optional): python logger object
        author (optional): dict of name, email, time - default uses auth'd info
        """
        self.repo = repo
        if not logger:
            logger = logging.getLogger(os.path.basename(__file__))
        self.logger = logger
        self.author = author if author else {}

    def __call__(self,
            local_dir,
            branch,
            repo_dir=None,
            commit_message=None,
            dry_run=False,
        ):
        """
        local_dir: path to local directory to commit
        branch: target branch name
        repo_dir: target path within repo - use '' for repo root
        commit_message: message for git commit
        dry_run: skip creating GitHub data if True
        """

        # prepare dir args
        local_dir = os.path.abspath(local_dir)
        if not os.path.isdir(local_dir):
            raise GithubException('Not a dir: {}'.format(local_dir))
        # do not use os.path because repo_dir is not local
        if repo_dir is None:
            repo_dir = ''
        if repo_dir.startswith('.'):
            repo_dir = repo_dir[1:]
        if repo_dir.startswith('/'):
            repo_dir = repo_dir[1:]
        if repo_dir.endswith('/'):
            repo_dir = repo_dir[:-1]

        # get ref to branch HEAD
        head = self.repo.ref('heads/{}'.format(branch))

        # get commit pointed to by HEAD
        commit = self.repo.git_commit(head.object.sha)

        # get tree of commit
        tree = self.repo.tree(commit.tree.sha) # shallow tree
        tree = tree.recurse() # recurse/flatten
        tree = tree.to_json() # convert to native types

        # create new tree (delete, update)
        new_tree_list = []
        for item in tree['tree']:
            if item['type'] == 'tree':
                # remove sub-trees as they are implied by blob paths
                self.logger.debug('Removing tree: {}'.format(item['path']))
                continue
            if not item['path'].startswith(repo_dir):
                # outside target dir in repo - keep in tree
                self.logger.debug(
                    'Unchanged (outside target path): {}'.format(item['path'])
                )
                new_tree_list.append(item)
                continue
            len_path = (len(repo_dir) + 1) if repo_dir else 0
            item_subpath = item['path'][len_path:]
            local_file = os.path.join(local_dir, item_subpath)
            if not os.path.isfile(local_file):
                # delete blob from tree
                self.logger.debug('Delete: {}'.format(item['path']))
                continue
            with io.open(local_file, 'rb') as f:
                content = f.read()
            header = 'blob {}\0'.format(len(content))
            sha = hashlib.sha1(header + content).hexdigest()
            new_item = item.copy()
            if sha != item['sha']:
                self.logger.debug('Update: {}'.format(local_file))
                if dry_run:
                    self.logger.info('[dry_run] Skipping creation of blob ' +
                        'for file: {}'.format(local_file)
                    )
                    blob_sha = None
                else:
                    self.logger.info(
                        'Creating blob for updated file: {}'.format(local_file)
                    )
                    blob_sha = self._create_blob(content)
                    self.logger.debug('Blob created: {}'.format(blob_sha))
                new_item['sha'] = blob_sha
            else:
                self.logger.debug('Unchanged: {}'.format(item['path']))
            new_tree_list.append(new_item)

        # add new files to tree
        new_tree_target_subpaths = []
        for item in new_tree_list:
            if not item['path'].startswith(repo_dir):
                # skip items not in target dir
                continue
            len_path = (len(repo_dir) + 1) if repo_dir else 0
            new_tree_target_subpaths.append(item['path'][len_path:])
        for root, dirs, files in os.walk(local_dir):
            for filename in files:
                if filename.startswith('.'):
                    # skip hidden files
                    continue
                local_file = os.path.join(root, filename)
                local_file_subpath = local_file[(len(local_dir) + 1):]
                if local_file_subpath not in new_tree_target_subpaths:
                    with io.open(local_file, 'rb') as f:
                        content = f.read()
                    if dry_run:
                        self.logger.info('[dry_run] Skipping creation of ' +
                            'blob for new file: {}'.format(local_file)
                        )
                        blob_sha = None
                    else:
                        self.logger.info(
                            'Creating blob for new file: {}'.format(local_file)
                        )
                        blob_sha = self._create_blob(content)
                        self.logger.debug('Blob created: {}'.format(blob_sha))
                    repo_path = (repo_dir + '/') if repo_dir else ''
                    new_item = {
                        'path': '{}{}'.format(repo_path, local_file_subpath),
                        'mode': '100644',
                        'type': 'blob',
                        'sha': blob_sha,
                    }
                    new_tree_list.append(new_item)

        # generate summary of changes
        self.logger.info('Summary of changes:')
        new_shas = [item['sha'] for item in new_tree_list]
        new_paths = [item['path'] for item in new_tree_list]
        old_paths = [item['path'] for item in tree['tree']]
        old_tree_list = []
        for item in tree['tree']:
            if item['type'] == 'tree':
                continue
            old_tree_list.append(item)
            if item['path'] not in new_paths:
                self.logger.warning('Delete:\t{}'.format(item['path']))
            elif item['sha'] not in new_shas:
                self.logger.info('Update:\t{}'.format(item['path']))
        for item in new_tree_list:
            if item['path'] not in old_paths:
                self.logger.info('Add:\t{}'.format(item['path']))
        if new_tree_list == old_tree_list:
            self.logger.warning('No changes found, aborting commit')
            return


        # create new tree
        if dry_run:
            self.logger.info('[dry_run] Skipping creation of new tree')
        else:
            self.logger.info('Creating new tree')
            new_tree = self.repo.create_tree(new_tree_list, None)
            if not new_tree:
                raise GithubException('Failed to create tree')

        # create new commit
        if commit_message is None:
            commit_message = 'Commit dir {} to {}/{}/{} via CumulusCI'.format(
                local_dir,
                self.repo.owner,
                self.repo.name,
                repo_dir,
            )
        if dry_run:
            self.logger.info('[dry_run] Skipping creation of new commit')
        else:
            self.logger.info('Creating new commit')
            new_commit = self.repo.create_commit(
                message=commit_message,
                tree=new_tree.sha,
                parents=[commit.sha],
                author=self.author,
                committer=self.author,
            )
            if not new_commit:
                raise GithubException('Failed to create commit')

        # update HEAD
        if dry_run:
            self.logger.info('[dry_run] Skipping call to update HEAD')
        else:
            self.logger.info('Updating HEAD')
            success = head.update(new_commit.sha)
            if not success:
                raise GithubException('Failed to update HEAD')

    def _create_blob(self, content):
        try:
            content = content.decode('utf-8')
            blob_sha = self.repo.create_blob(content, 'utf-8')
        except UnicodeDecodeError:
            content = base64.b64encode(content)
            blob_sha = self.repo.create_blob(content, 'base64')
        if not blob_sha:
            raise GithubException('Failed to create blob')
        return blob_sha



import os

from cumulusci.tasks.salesforce import Deploy


deploy_options = Deploy.task_options.copy()
deploy_options['path']['description'] = 'The path to the parent directory containing the metadata bundles directories'
class DeployBundles(Deploy):
    task_options = deploy_options

    def _run_task(self):
        path = self.options['path']
        pwd = os.getcwd()

        path = os.path.join(pwd, path)

        self.logger.info(
            'Deploying all metadata bundles in path {}'.format(path))

        if not os.path.isdir(path):
            self.logger.warn('Path {} not found, skipping'.format(path))
            return

        for item in sorted(os.listdir(path)):
            item_path = os.path.join(path, item)
            if not os.path.isdir(item_path):
                continue

            self.logger.info('Deploying bundle: {}/{}'.format(self.options['path'], item))

            self._deploy_bundle(item_path)

    def _deploy_bundle(self, path):
        api = self._get_api(path)
        return api()



""" a task for waiting on a Batch Apex job to complete """

from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.exceptions import SalesforceException
import arrow

COMPLETED_STATUSES = ['Completed']

class BatchApexWait(BaseSalesforceApiTask):
    """ BatchApexWait polls an org until the latest batch job
        for an apex class completes or fails """

    name = 'BatchApexWait'
    batch = object()

    task_options = {
        'class_name': {
            'description': 'Name of the Apex class to wait for.',
            'required': True
        },
        'poll_interval': {
            'description': 'Seconds to wait before polling for batch job completion. ' \
            'Defaults to 10 seconds.'
        }
    }

    def _run_task(self):
        self.poll_interval_s = int(self.options.get('poll_interval', 10))

        self._poll() # will block until poll_complete

        self.logger.info('Job is complete.')

        if not self.success:
            self.logger.info('There were some batch failures.')
            raise SalesforceException(self.batch['ExtendedStatus'])

        self.logger.info('%s took %d seconds to process %d batches.',
                         self.batch['ApexClass']['Name'],
                         self.delta,
                         self.batch['TotalJobItems'])

        return self.success

    def _poll_action(self):
        # get batch status
        query_results = self.tooling.query(self._batch_query)

        self.batch = query_results['records'][0]
        self.logger.info('%s: %d of %d (%d failures)',
                         self.batch['ApexClass']['Name'],
                         self.batch['JobItemsProcessed'],
                         self.batch['TotalJobItems'],
                         self.batch['NumberOfErrors'])

        self.poll_complete = not self._poll_again()

    def _poll_again(self):
        return self.batch['Status'] not in COMPLETED_STATUSES

    @property
    def success(self):
        """ returns True if all batches succeeded """
        return (self.batch['JobItemsProcessed'] is self.batch['TotalJobItems']) and \
               (self.batch['NumberOfErrors'] is 0)

    @property
    def delta(self):
        """ returns the time (in seconds) that the batch took, if complete """
        td = arrow.get(self.batch['CompletedDate']) - \
            arrow.get(self.batch['CreatedDate'])

        return td.total_seconds()

    @property
    def _batch_query(self):
        return 'SELECT Id, ApexClass.Name, Status, ExtendedStatus, TotalJobItems, ' \
                  'JobItemsProcessed, NumberOfErrors, CreatedDate, CompletedDate ' \
                  'FROM AsyncApexJob ' \
                  'WHERE JobType=\'BatchApex\' '\
                  'AND ApexClass.Name=\'{}\' ' \
                  'ORDER BY CreatedDate DESC ' \
                  'LIMIT 1'.format(self.options['class_name'])



""" Tasks are the basic unit of execution in CumulusCI.

Subclass BaseTask or a descendant to define custom task logic
"""
from __future__ import division
from __future__ import unicode_literals

from builtins import object
from past.utils import old_div
import logging
import time

from cumulusci.core.exceptions import TaskRequiresSalesforceOrg
from cumulusci.core.exceptions import TaskOptionsError


class BaseTask(object):
    """ BaseTask provides the core execution logic for a Task

    Subclass BaseTask and provide a `_run_task()` method with your
    code.
    """
    task_options = {}
    salesforce_task = False  # Does this task require a salesforce org?

    def __init__(
        self,
        project_config,
        task_config,
        org_config=None,
        flow=None,
        name=None,
        stepnum=None,
        **kwargs
    ):
        self.project_config = project_config
        self.task_config = task_config
        self.org_config = org_config
        self.poll_count = 0
        self.poll_interval_level = 0
        self.poll_interval_s = 1
        self.poll_complete = False
        

        # dict of return_values that can be used by task callers
        self.return_values = {}

        # simple result object for introspection, often a return_code
        self.result = None

        # the flow for this task execution
        self.flow = flow
        
        # the tasks name in the flow
        self.name = name

        # the tasks stepnumber in the flow
        self.stepnum = stepnum

        if self.salesforce_task and not self.org_config:
            raise TaskRequiresSalesforceOrg(
                'This task requires a Saleforce org_config but' +
                ' none was passed to the Task constructor'
            )
        self._init_logger()
        self._init_options(kwargs)
        self._validate_options()
        self._update_credentials()
        self._init_task()

    def _init_logger(self):
        """ Initializes self.logger """
        self.logger = logging.getLogger(__name__)

    def _init_options(self, kwargs):
        """ Initializes self.options """
        self.options = self.task_config.options
        if self.options is None:
            self.options = {}
        if kwargs:
            self.options.update(kwargs)

        # Handle dynamic lookup of project_config values via $project_config.attr
        for option, value in list(self.options.items()):
            try:
                if value.startswith('$project_config.'):
                    attr = value.replace('$project_config.', '', 1)
                    self.options[option] = getattr(self.project_config, attr, None)
            except AttributeError:
                pass
                

    def _validate_options(self):
        missing_required = []
        for name, config in list(self.task_options.items()):
            if config.get('required') is True and name not in self.options:
                missing_required.append(name)

        if missing_required:
            raise TaskOptionsError(
                '{} requires the options ({}) '
                'and no values were provided'.format(
                    self.__class__.__name__,
                    ', '.join(missing_required),
                )
            )

    def _update_credentials(self):
        """ Override to do any logic  to refresh credentials """
        pass

    def _init_task(self):
        """ Override to implement dynamic logic for initializing the task. """
        pass

    def __call__(self):
        # If sentry is configured, initialize sentry for error capture
        self.project_config.init_sentry()

        try:
            self._log_begin()
            self.result = self._run_task()
            return self.return_values
        except Exception as e:
            self._process_exception(e)
            raise

    def _process_exception(self, e):
        if self.project_config.use_sentry:
            self.logger.info('Logging error to sentry.io')

            tags = {
                'task class': self.__class__.__name__,
            }
            if self.org_config:
                tags['org username'] = self.org_config.username
                tags['scratch org'] = self.org_config.scratch == True
            for key, value in list(self.options.items()):
                tags['option_' + key] = value
            self.project_config.sentry.tags_context(tags)

            resp = self.project_config.sentry.captureException()
            self.project_config.sentry_event = resp

    def _run_task(self):
        """ Subclasses should override to provide their implementation """
        pass

    def _log_begin(self):
        """ Log the beginning of the task execution """
        self.logger.info('Beginning task: %s', self.__class__.__name__)
        if self.salesforce_task and not self.flow:
            self.logger.info('%15s %s', 'As user:', self.org_config.username)
            self.logger.info('%15s %s', 'In org:', self.org_config.org_id)
        self.logger.info('')

    def _retry(self):
        while True:
            try:
                self._try()
                break
            except Exception as e:
                if not (self.options['retries'] and self._is_retry_valid(e)):
                    raise
                if self.options['retry_interval']:
                    self.logger.warning(
                        'Sleeping for {} seconds before retry...'.format(
                            self.options['retry_interval']
                        )
                    )
                    time.sleep(self.options['retry_interval'])
                    if self.options['retry_interval_add']:
                        self.options['retry_interval'] += (
                            self.options['retry_interval_add']
                        )
                self.options['retries'] -= 1
                self.logger.warning(
                    'Retrying ({} attempts remaining)'.format(
                        self.options['retries']
                    )
                )

    def _try(self):
        raise NotImplementedError(
            'Subclasses should provide their own implementation'
        )

    def _is_retry_valid(self, e):
        return True

    def _poll(self):
        ''' poll for a result in a loop '''
        while True:
            self.poll_count += 1
            self._poll_action()
            if self.poll_complete:
                break
            time.sleep(self.poll_interval_s)
            self._poll_update_interval()

    def _poll_action(self):
        '''
        Poll something and process the response.
        Set `self.poll_complete = True` to break polling loop.
        '''
        raise NotImplementedError(
            'Subclasses should provide their own implementation'
        )

    def _poll_update_interval(self):
        ''' update the polling interval to be used next iteration '''
        # Increase by 1 second every 3 polls
        if old_div(self.poll_count, 3) > self.poll_interval_level:
            self.poll_interval_level += 1
            self.poll_interval_s += 1
            self.logger.info(
                'Increased polling interval to %d seconds',
                self.poll_interval_s,
            )



from cumulusci.core.exceptions import ApexTestException
from cumulusci.core.exceptions import SalesforceException
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class PackageUpload(BaseSalesforceApiTask):
    name = 'PackageUpload'
    api_version = '38.0'
    task_options = {
        'name': {
            'description': 'The name of the package version.',
            'required': True,
        },
        'production': {
            'description': 'If True, uploads a production release.  Defaults to uploading a beta',
        },
        'description': {
            'description': 'A description of the package and what this version contains.',
        },
        'password': {
            'description': "An optional password for sharing the package privately with anyone who has the password. Don't enter a password if you want to make the package available to anyone on AppExchange and share your package publicly.",
        },
        'post_install_url': {
            'description': 'The fully-qualified URL of the post-installation instructions. Instructions are shown as a link after installation and are available from the package detail view.',
        },
        'release_notes_url': {
            'description': 'The fully-qualified URL of the package release notes. Release notes are shown as a link during the installation process and are available from the package detail view after installation.',
        },
        'namespace': {
            'description': 'The namespace of the package.  Defaults to project__package__namespace',
        },
    }

    def _init_options(self, kwargs):
        super(PackageUpload, self)._init_options(kwargs)

        self.upload = None
        self.upload_id = None

        # Set the namespace option to the value from cumulusci.yml if not already set
        if not 'namespace' in self.options:
            self.options['namespace'] = self.project_config.project__package__namespace

    def _run_task(self):
        package_res = self.tooling.query(
            "select Id from MetadataPackage where NamespacePrefix='{}'".format(self.options['namespace']))

        if package_res['totalSize'] != 1:
            message = 'No package found with namespace {}'.format(
                self.options['namespace'])
            self.logger.error(message)
            raise SalesforceException(message)

        package_id = package_res['records'][0]['Id']

        production = self.options.get('production', False) in [
            True, 'True', 'true']
        package_info = {
            'VersionName': self.options['name'],
            'IsReleaseVersion': production,
            'MetadataPackageId': package_id,
        }

        if 'description' in self.options:
            package_info['Description'] = self.options['description']
        if 'password' in self.options:
            package_info['Password'] = self.options['password']
        if 'post_install_url' in self.options:
            package_info['PostInstallUrl'] = self.options['post_install_url']
        if 'release_notes_url' in self.options:
            package_info['ReleaseNotesUrl'] = self.options['release_notes_url']

        PackageUploadRequest = self._get_tooling_object('PackageUploadRequest')
        self.upload = PackageUploadRequest.create(package_info)
        self.upload_id = self.upload['id']

        self.logger.info('Created PackageUploadRequest {} for Package {}'.format(
            self.upload_id, package_id))
        self._poll()

        if self.upload['Status'] == 'ERROR':
            self.logger.error(
                'Package upload failed with the following errors')
            for error in self.upload['Errors']['errors']:
                self.logger.error('  {}'.format(error['message']))

            # use the last error in the batch, but log them all.
            if error['message'] == 'ApexTestFailure':
                e = ApexTestException
            else:
                e = SalesforceException
            raise e('Package upload failed')
        else:
            version_id = self.upload['MetadataPackageVersionId']
            version_res = self.tooling.query(
                "select MajorVersion, MinorVersion, PatchVersion, BuildNumber, ReleaseState from MetadataPackageVersion where Id = '{}'".format(version_id))
            if version_res['totalSize'] != 1:
                message = 'Version {} not found'.format(version_id)
                self.logger.error(message)
                raise SalesforceException(message)

            version = version_res['records'][0]
            version_parts = [
                str(version['MajorVersion']),
                str(version['MinorVersion']),
            ]
            if version['PatchVersion']:
                version_parts.append(str(version['PatchVersion']))

            self.version_number = '.'.join(version_parts)

            if version['ReleaseState'] == 'Beta':
                self.version_number += ' (Beta {})'.format(
                    version['BuildNumber'])

            self.return_values = {
                'version_number': str(self.version_number),
                'version_id': version_id,
                'package_id': package_id
            }

            self.logger.info('Uploaded package version {} with Id {}'.format(
                self.version_number,
                version_id
            ))

    def _poll_action(self,):
        soql_check_upload = "select Id, Status, Errors, MetadataPackageVersionId from PackageUploadRequest where Id = '{}'".format(
            self.upload_id)

        uploadresult = self.tooling.query(soql_check_upload)
        if uploadresult['totalSize'] != 1:
            message = 'Failed to get info for upload with id {}'.format(
                self.upload_id)
            self.logger.error(message)
            raise SalesforceException(message)

        self.upload = uploadresult['records'][0]
        self.logger.info('PackageUploadRequest {} is {}'.format(
            self.upload_id, self.upload['Status']))

        self.poll_complete = not self._poll_again(self.upload['Status'])

    def _poll_again(self, upload_status):
        return upload_status in ['IN_PROGRESS', 'QUEUED']



from __future__ import unicode_literals
import os

import hiyapyco
import yaml

from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.exceptions import NotInProject
from cumulusci.core.exceptions import ProjectConfigNotFound


class YamlProjectConfig(BaseProjectConfig):
    config_filename = 'cumulusci.yml'

    def __init__(self, *args, **kwargs):
        # Initialize the dictionaries for the individual configs
        self.config_project = {}
        self.config_project_local = {}
        self.config_additional_yaml = {}

        # optionally pass in a kwarg named 'additional_yaml' that will
        # be added to the YAML merge stack.
        self.additional_yaml = None
        if 'additional_yaml' in kwargs:
            self.additional_yaml = kwargs.pop('additional_yaml')

        super(YamlProjectConfig, self).__init__(*args, **kwargs)

    @property
    def config_project_local_path(self):
        path = os.path.join(
            self.project_local_dir,
            self.config_filename,
        )
        if os.path.isfile(path):
            return path

    def _load_config(self):
        """ Loads the configuration for the project """
        # Verify that we're in a project
        repo_root = self.repo_root
        if not repo_root:
            raise NotInProject(
                'No repository found in current path.  You must be inside a repository to initialize the project configuration')

        # Verify that the project's root has a config file
        if not self.config_project_path:
            raise ProjectConfigNotFound(
                'The file {} was not found in the repo root: {}'.format(
                    self.config_filename,
                    repo_root
                )
            )

        # Start the merged yaml config from the global and global local configs
        merge_yaml = [self.global_config_obj.config_global_path]
        if self.global_config_obj.config_global_local_path:
            merge_yaml.append(self.global_config_obj.config_global_local_path)

        # Load the project's yaml config file
        with open(self.config_project_path, 'r') as f_config:
            project_config = yaml.load(f_config)
        if project_config:
            self.config_project.update(project_config)
            merge_yaml.append(self.config_project_path)

        # Load the local project yaml config file if it exists
        if self.config_project_local_path:
            with open(self.config_project_local_path, 'r') as f_local_config:
                local_config = yaml.load(f_local_config)
            if local_config:
                self.config_project_local.update(local_config)
                merge_yaml.append(self.config_project_local_path)

        # merge in any additional yaml that was passed along
        if self.additional_yaml:
            additional_yaml_config = yaml.load(self.additional_yaml)
            if additional_yaml_config:
                self.config_additional_yaml.update(additional_yaml_config)
                merge_yaml.append(self.additional_yaml)

        self.config = hiyapyco.load(
            *merge_yaml,
            method=hiyapyco.METHOD_MERGE,
            loglevel='INFO'
        )



""" Tests for the Flow engine """

import unittest
import logging
import mock

from collections import Callable
from nose.tools import raises

from cumulusci.core.flows import BaseFlow
from cumulusci.core.tasks import BaseTask
from cumulusci.core.config import FlowConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.exceptions import FlowConfigError
from cumulusci.core.exceptions import TaskNotFoundError
from cumulusci.core.tests.utils import MockLoggingHandler
from cumulusci.tests.util import create_project_config
import cumulusci.core

ORG_ID = "00D000000000001"


class _TaskReturnsStuff(BaseTask):

    def _run_task(self):
        self.return_values = {'name': 'supername'}


class _TaskResponseName(BaseTask):
    task_options = {'response': {'description': 'the response to print'}}

    def _run_task(self):
        return self.options['response']

class _TaskRaisesException(BaseTask):
    task_options = {
        'exception': {'description': 'The exception to raise'},
        'message': {'description': 'The exception message'},
    }

    def _run_task(self):
        raise self.options['exception'](self.options['message'])

class _SfdcTask(BaseTask):
    salesforce_task = True

    def _run_task(self):
        return -1

@mock.patch('cumulusci.core.flows.BaseFlow._init_org')
class TestBaseFlow(unittest.TestCase):
    """ Tests the expectations of a BaseFlow caller """

    @classmethod
    def setUpClass(cls):
        super(TestBaseFlow, cls).setUpClass()
        logger = logging.getLogger(cumulusci.core.__name__)
        logger.setLevel(logging.DEBUG)
        cls._flow_log_handler = MockLoggingHandler(logging.DEBUG)
        logger.addHandler(cls._flow_log_handler)

    def setUp(self):
        self.project_config = create_project_config('TestOwner', 'TestRepo')
        self.project_config.config['tasks'] = {
            'pass_name': {
                'description': 'Pass the name',
                'class_path':
                    'cumulusci.core.tests.test_flows._TaskReturnsStuff',
            },
            'name_response': {
                'description': 'Pass the name',
                'class_path':
                    'cumulusci.core.tests.test_flows._TaskResponseName',
            },
            'raise_exception': {
                'description': 'Raises an exception',
                'class_path':
                    'cumulusci.core.tests.test_flows._TaskRaisesException',
                'options': {
                    'exception': Exception,
                    'message': 'Test raised exception as expected',
                }
            },
            'sfdc_task': {
                'description': 'An sfdc task',
                'class_path':
                    'cumulusci.core.tests.test_flows._SfdcTask'
            },
        }
        self.project_config.config['flows'] = {
            'nested_flow': {
                'description': 'A flow that runs inside another flow',
                'steps': {
                    1: {'task': 'pass_name'},
                },
            },
            'nested_flow_2': {
                'description': 'A flow that runs inside another flow, and calls another flow',
                'steps': {
                    1: {'task': 'pass_name'},
                    2: {'flow': 'nested_flow'},
                },
            },
        }
        self.org_config = OrgConfig({
            'username': 'sample@example',
            'org_id': ORG_ID
        }, 'test')

        self._flow_log_handler.reset()
        self.flow_log = self._flow_log_handler.messages

    def test_init(self, mock_class):
        """ BaseFlow initializes and offers a logger """
        flow_config = FlowConfig({})
        mock_class.return_value = None
        flow = BaseFlow(self.project_config, flow_config, self.org_config)

        self.assertEquals(hasattr(flow, 'logger'), True)

    def test_is_callable(self, mock_class):
        """ BaseFlow exposes itself as a callable for use """
        flow_config = FlowConfig({})
        flow = BaseFlow(self.project_config, flow_config, self.org_config)

        self.assertIsInstance(flow, Callable)

    def test_pass_around_values(self, mock_class):
        """ A flow's options reach into return values from other tasks. """

        mock_class.return_value = None
        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'name_response', 'options': {
                    'response': '^^pass_name.name'
                }},
            }
        })

        flow = BaseFlow(self.project_config, flow_config, self.org_config)

        # run the flow
        flow()
        # the flow results for the second task should be 'name'
        self.assertEquals('supername', flow.step_results[1])

    def test_task_options(self, mock_class):
        """ A flow can accept task options and pass them to the task. """

        mock_class.return_value = None
        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'name_response', 'options': {
                    'response': 'foo'
                }},
            }
        })

        flow = BaseFlow(
            self.project_config,
            flow_config,
            self.org_config,
            options={'name_response__response': 'bar'},
        )

        # run the flow
        flow()
        # the flow results for the first task should be 'bar'
        self.assertEquals('bar', flow.step_results[0])

    def test_skip_kwarg(self, mock_class):
        """ A flow can receive during init a list of tasks to skip """

        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'name_response', 'options': {
                    'response': '^^pass_name.name'
                }},
            }
        })

        flow = BaseFlow(
            self.project_config,
            flow_config,
            self.org_config,
            skip=['name_response'],
        )

        # run the flow
        flow()

        # the number of tasks in the flow should be 1 instead of 2
        self.assertEquals(1, len(flow.step_results))

    def test_skip_task_value_none(self, mock_class):
        """ A flow skips any tasks whose name is None to allow override via yaml """

        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'None'},
            }
        })

        flow = BaseFlow(
            self.project_config,
            flow_config,
            self.org_config,
            skip=['name_response'],
        )

        # run the flow
        flow()

        # the number of tasks in the flow should be 1 instead of 2
        self.assertEquals(1, len(flow.step_results))

    @raises(FlowConfigError)
    def test_find_step_by_name_no_steps(self, mock_class):
        """ Running a flow with no steps throws an error """

        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run two tasks',
        })

        flow = BaseFlow(
            self.project_config,
            flow_config,
            self.org_config,
        )
        flow()

    def test_find_step_by_name_not_first(self, mock_class):
        """ The _find_step_by_name method skips tasks that don't exist """

        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'name_response', 'options': {
                    'response': '^^pass_name.name'
                }},
            }
        })

        flow = BaseFlow(
            self.project_config,
            flow_config,
            self.org_config,
        )

        flow()

        task = flow._find_step_by_name('name_response')
        self.assertEquals(
            'cumulusci.core.tests.test_flows._TaskResponseName',
            task.task_config.class_path,
        )

    def test_render_task_config_empty_value(self, mock_class):
        """ The _render_task_config method skips option values of None """

        # instantiate a flow with two tasks
        flow_config = FlowConfig({
            'description': 'Run a tasks',
            'steps': {
                1: {'task': 'name_response',
                    'options': {
                        'response': None,
                    },
            
                   },
            },
        })

        flow = BaseFlow(
            self.project_config,
            flow_config,
            self.org_config,
        )

        flow()

        task = flow._find_step_by_name('name_response')
        config = flow._render_task_config(task)
        self.assertEquals(['Options:'], config)

    def test_task_raises_exception_fail(self, mock_class):
        """ A flow aborts when a task raises an exception """

        flow_config = FlowConfig({
            'description': 'Run a task',
            'steps': {
                1: {'task': 'raise_exception'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        self.assertRaises(Exception, flow)

    def test_task_raises_exception_ignore(self, mock_class):
        """ A flow continues when a task configured with ignore_failure raises an exception """

        flow_config = FlowConfig({
            'description': 'Run a task',
            'steps': {
                1: {'task': 'raise_exception', 'ignore_failure': True},
                2: {'task': 'pass_name'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()
        self.assertEquals(2, len(flow.steps))

    def test_call_no_tasks(self, mock_class):
        """ A flow with no tasks will have no responses. """
        flow_config = FlowConfig({
            'description': 'Run no tasks',
            'steps': {}
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()

        self.assertEqual([], flow.step_return_values)
        self.assertEqual([], flow.steps)

    def test_call_one_task(self, mock_class):
        """ A flow with one task will execute the task """
        flow_config = FlowConfig({
            'description': 'Run one task',
            'steps': {
                1: {'task': 'pass_name'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()

        self.assertTrue(any(
            "Flow Description: Run one task" in s for s in self.flow_log['info']
        ))

        self.assertEqual([{'name': 'supername'}], flow.step_return_values)
        self.assertEqual(1, len(flow.steps))

    def test_call_many_tasks(self, mock_class):
        """ A flow with many tasks will dispatch each task """
        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'pass_name'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()

        self.assertEqual(
            [{'name': 'supername'}, {'name': 'supername'}],
            flow.step_return_values
        )
        self.assertEqual(2, len(flow.steps))

    @raises(TaskNotFoundError)
    def test_call_task_not_found(self, mock_class):
        """ A flow with reference to a task that doesn't exist in the
        project will throw a TaskNotFoundError """

        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'do_delightulthings'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)

    def test_flow_prints_org_id(self, mock_class):
        """ A flow with an org prints the org ID """

        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'pass_name'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()

        org_id_logs = [s for s in self.flow_log['info'] if ORG_ID in s]

        self.assertEqual(1, len(org_id_logs))

    def test_flow_no_org_no_org_id(self, mock_class):
        """ A flow without an org does not print the org ID """

        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'task': 'pass_name'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, None)
        flow()

        self.assertFalse(any(
            ORG_ID in s for s in self.flow_log['info']
        ))

    def test_flow_prints_org_id_once_only(self, mock_class):
        """ A flow with sf tasks prints the org ID only once."""

        flow_config = FlowConfig({
            'description': 'Run two tasks',
            'steps': {
                1: {'task': 'sfdc_task'},
                2: {'task': 'sfdc_task'},
            }
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()

        org_id_logs = [s for s in self.flow_log['info'] if ORG_ID in s]

        self.assertEqual(1, len(org_id_logs))

    def test_nested_flow(self, mock_class):
        """ Flows can run inside other flows """
        flow_config = FlowConfig({
            'description': 'Run a task and a flow',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'flow': 'nested_flow'},
            },
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()
        self.assertEqual(2, len(flow.steps))
        self.assertEqual(
            flow.step_return_values[0],
            flow.step_return_values[1][0],
        )

    def test_nested_flow_2(self, mock_class):
        """ Flows can run inside other flows and call other flows """
        flow_config = FlowConfig({
            'description': 'Run a task and a flow',
            'steps': {
                1: {'task': 'pass_name'},
                2: {'flow': 'nested_flow_2'},
            },
        })
        flow = BaseFlow(self.project_config, flow_config, self.org_config)
        flow()
        self.assertEqual(2, len(flow.steps))
        self.assertEqual(
            flow.step_return_values[0],
            flow.step_return_values[1][0],
        )
        self.assertEqual(
            flow.step_return_values[0],
            flow.step_return_values[1][1][0],
        )



from cumulusci.tasks.salesforce import Deploy
from cumulusci.salesforce_api.package_zip import CreatePackageZipBuilder


class CreatePackage(Deploy):
    task_options = {
        'package': {
            'description': 'The name of the package to create.  Defaults to project__package__name',
            'required': True,
        },
        'api_version': {
            'description': 'The api version to use when creating the package.  Defaults to project__package__api_version',
            'required': True,
        },
    }

    def _init_options(self, kwargs):
        super(CreatePackage, self)._init_options(kwargs)
        if 'package' not in self.options:
            self.options['package'] = self.project_config.project__package__name
        if 'api_version' not in self.options:
            self.options['api_version'] = self.project_config.project__package__api_version

    def _get_api(self, path=None):
        package_zip = CreatePackageZipBuilder(self.options['package'], self.options['api_version'])
        return self.api_class(self, package_zip(), purge_on_delete=False)



""" The flow engine allows a series of tasks to be run. """
from __future__ import unicode_literals

from builtins import str
from builtins import object
import copy
from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module
import logging
import traceback

from cumulusci.core.config import FlowConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import FlowConfigError
from cumulusci.core.exceptions import FlowInfiniteLoopError
from cumulusci.core.exceptions import FlowNotReadyError
from cumulusci.core.utils import import_class


class BaseFlow(object):
    """ BaseFlow handles initializing and running a flow """

    def __init__(
            self,
            project_config,
            flow_config,
            org_config,
            options=None,
            skip=None,
            nested=False,
            parent=None,
            prep=True,
            name=None,
            stepnum=None
    ):
        self.project_config = project_config # a subclass of BaseTaskFlowConfig, tho tasks may expect more than that
        self.flow_config = flow_config
        self.org_config = org_config
        self.options = options
        self.skip = skip
        self.task_options = {}
        self.skip_steps = []
        self.step_return_values = []  # A collection of return_values dicts in task execution order
        self.step_results = []  # A collection of result objects in task execution order
        self.steps = []  # A collection of configured task objects, either run or failed
        self.nested = nested  # indicates if flow is called from another flow
        self.parent = parent # parent flow, if nested
        self.name = name # the flows name.
        self.stepnum = stepnum # a nested flow has a stepnum
        self._skip_next = False # internal only control flow option for subclasses to override a step execution in their pre.
        self._init_options()
        self._init_skip(skip)
        self._init_logger()
        self.prepped = False
        if prep:
            self._init_flow()

    def _init_options(self):
        if not self.options:
            return
        for key, value in list(self.options.items()):
            task, option = key.split('__')
            if task not in self.task_options:
                self.task_options[task] = {}
            self.task_options[task][option] = value

    def _init_skip(self, skip):
        if not skip:
            return
        for step in skip:
            self.skip_steps.append(step)

    def _init_logger(self):
        """ Initializes self.logger """
        self.logger = logging.getLogger(__name__)

    def _init_org(self):
        """ Refresh the token on the org """
        self.logger.info('Verifying and refreshing credentials for target org {}'.format(self.org_config.name))
        orig_config = self.org_config.config.copy()
        self.org_config.refresh_oauth_token(self.project_config.keychain)
        if self.org_config.config != orig_config:
            self.logger.info('Org info has changed, updating org in keychain')
            self.project_config.keychain.set_org(self.org_config)

    def _init_flow(self):
        """ Initialize the flow and print flow details to info """
        self.logger.info('---------------------------------------')
        if self.name:
            self.logger.info(
                'Initializing flow: {}'.format(self.name)
            )
        else:
            self.logger.info(
                'Initializing flow class: {}'.format(self.__class__.__name__)
            )
        self.logger.info('---------------------------------------')

        self.logger.info('')
        if not self.nested:
            self._init_org()
        for line in self._render_config():
            self.logger.info(line)
        self.prepped = True

    def _check_infinite_flows(self, steps, flows=None):
        if flows == None:
            flows = []
        for step in steps.values():
            if 'flow' in step:
                flow = step['flow']
                if flow in flows:
                    raise FlowInfiniteLoopError('Infinite flows detected with flow {}'.format(flow))
                flows.append(flow)
                flow_config = self.project_config.get_flow(flow)
                self._check_infinite_flows(flow_config.steps, flows)

    def _get_steps(self):
        steps = self._get_steps_ordered()
        steps = [step[1] for step in steps]  # drop the sort keys
        return steps

    def _get_steps_ordered(self):
        if self.flow_config.steps is None:
            if self.flow_config.tasks:
                raise FlowConfigError('Old flow syntax detected.  Please change from "tasks" to "steps" in the flow definition')
            else:
                raise FlowConfigError('No steps found in the flow definition')
        if not self.nested:
            self._check_infinite_flows(self.flow_config.steps)
        steps = []
        for step_num, config in list(self.flow_config.steps.items()):
            if 'flow' in config and 'task' in config:
                raise FlowConfigError('"flow" and "task" in same config item: {}'.format(config))
            if (('flow' in config and config['flow'] == 'None') or
                ('task' in config and config['task'] == 'None')):
                # allows skipping flows/tasks using YAML overrides
                continue
            parsed_step_num = LooseVersion(str(step_num))
            if 'flow' in config:  # nested flow
                flow_config = self.project_config.get_flow(config['flow'])
                steps.append((
                    parsed_step_num,
                    {
                        'step_config': config,
                        'flow_config': flow_config,
                    },
                ))
            elif 'task' in config:
                task_config = self.project_config.get_task(config['task'])
                steps.append((
                    parsed_step_num,
                    {
                        'step_config': config,
                        'task_config': task_config,
                    },
                ))
        steps.sort()  # sort by step number
        return steps

    def _render_config(self):
        config = []
        config.append(
            'Flow Description: {}'.format(self.flow_config.description)
        )

        if not self.flow_config.steps:
            return config

        config.append('Steps:')
        for step_config in self._get_steps():
            if 'flow_config' in step_config:
                step = step_config['step_config']['flow']
                flow_prefix = '(Flow) '
                description = step_config['flow_config'].description
            elif 'task_config' in step_config:
                step = step_config['step_config']['task']
                flow_prefix = ''
                description = step_config['task_config'].description

            config.append('  {}{}{}: {}'.format(
                '[SKIPPED] ' if step in self.skip_steps else '',
                flow_prefix,
                step,
                description,
            ))

        if self.org_config is not None:
            config.append('Organization:')
            config.append('  {}: {}'.format(
                'Username', self.org_config.username))
            config.append('  {}: {}'.format(
                '  Org Id', self.org_config.org_id))

        return config

    def __call__(self):
        if not self.nested:
            self._pre_flow()
        if not self.prepped:
            raise FlowNotReadyError('Flow executed before init_flow was called')
        for stepnum, step_config in self._get_steps_ordered():
            self._run_step(stepnum, step_config)
        if not self.nested:
            self._post_flow()

    def _pre_flow(self):
        pass
    
    def _post_flow(self):
        pass

    def _find_step_by_name(self, name):
        if not self.flow_config.steps:
            return

        i = 0
        for step in self._get_steps():
            if 'flow_config' in step:
                item = step['step_config']['flow']
            elif 'task_config' in step:
                item = step['step_config']['task']
            if item == name:
                if len(self.steps) > i:
                    return self.steps[i]
            i += 1

    def _run_step(self, stepnum, step_config):
        if 'flow_config' in step_config:
            self._run_flow(stepnum, step_config)
        elif 'task_config' in step_config:
            self._run_task(stepnum, step_config)

    def _run_flow(self, stepnum, step_config):
        class_path = step_config['flow_config'].config.get(
            'class_path',
            'cumulusci.core.flows.BaseFlow',
        )
        flow_options = step_config['step_config'].get(
            'options',
            {},
        )
        if flow_options:
            # Collapse down flow options into task__option format to pass
            options = {}
            for task, task_options in flow_options.items():
                for option, value in task_options.items():
                    options['{}__{}'.format(task, option)] = value
            flow_options = options
                
        flow_class = import_class(class_path)
        flow = flow_class(
            self.project_config,
            step_config['flow_config'],
            self.org_config,
            options=flow_options,
            skip=self.skip,
            nested=True,
            parent=self,
            name=step_config['step_config']['flow'],
            stepnum=stepnum
        )
        self._pre_subflow(flow)
        
        if self._skip_next:
            self._skip_next = False
            return

        flow()
        self._post_subflow(flow)
        self.steps.append(flow)
        self.step_return_values.append(flow.step_return_values)

    def _run_task(self, stepnum, step_config):

        task = self._get_task(stepnum, step_config)

        # Skip the task if skip was requested
        if task.name in self.skip_steps:
            self.logger.info('')
            self.logger.info('Skipping task {}'.format(task.name))
            return

        self.steps.append(task)

        for line in self._render_task_config(task):
            self.logger.info(line)

        # NOW TRY TO RUN THE TASK
        self.logger.info('')
        self.logger.info('Running task: %s', task.name)
        self._pre_task(task)
        if self._skip_next:
            self._skip_next = False
            return
        try:
            task()
            self.logger.info('Task complete: %s', task.name)
            self.step_results.append(task.result)
            self.step_return_values.append(task.return_values)
            self._post_task(task)
        except Exception as e:
            self._post_task_exception(task, e)
            self.logger.error('Task failed: %s', task.name)
            if not step_config['step_config'].get('ignore_failure'):
                self.logger.error('Failing flow due to exception in task')
                traceback.print_exc()
                raise e
            self.logger.info('Continuing flow')

    def _pre_task(self, task):
        pass

    def _post_task(self, task):
        pass

    def _post_task_exception(self, task, exception):
        pass

    def _pre_subflow(self, flow):
        pass

    def _post_subflow(self, flow):
        pass

    def _get_task(self, stepnum, step_config):
        task_config = copy.deepcopy(step_config['task_config'].config)
        task_config = TaskConfig(task_config)

        task_name = step_config['step_config']['task']

        if 'options' not in task_config.config:
            task_config.config['options'] = {}
        task_config.config['options'].update(
            step_config['step_config'].get('options', {})
        )
        # If there were task option overrides passed in, merge them
        if task_name in self.task_options:
            task_config.config['options'].update(
                self.task_options[task_name]
            )

        # Handle dynamic value lookups in the format ^^task_name.attr1.attr2
        for option, value in list(task_config.options.items()):
            if str(value).startswith('^^'):
                value_parts = value[2:].split('.')
                parent = self._find_step_by_name(value_parts[0])
                n = 0
                while isinstance(parent, BaseFlow):
                    n += 1
                    parent = parent._find_step_by_name(value_parts[n])
                for attr in value_parts[(n + 1):]:
                    if getattr(parent, 'nested', None):
                        parent = parent._find_step_by_name()
                    else:
                        parent = parent.return_values.get(attr)
                task_config.config['options'][option] = parent

        task_class = import_class(task_config.class_path)

        task = task_class(
            self.project_config,
            task_config,
            org_config=self.org_config,
            name = task_name,
            stepnum=stepnum,
            flow=self
        )
        return task


    def _render_task_config(self, task):
        config = ['Options:']
        if not task.task_options:
            return config
        for option, info in list(task.task_options.items()):
            value = task.options.get(option)
            if value is None:
                continue
            config.append('  {}: {}'.format(
                option,
                value,
            ))

        return config



import httplib

import responses


class MockUtil(object):
    BASE_HTML_URL = 'https://github.com'
    BASE_API_URL = 'https://api.github.com'

    def __init__(self, owner, repo):
        self.owner = owner
        self.repo = repo

    @property
    def html_url(self):
        return '{}/{}/{}'.format(self.BASE_HTML_URL, self.owner, self.repo)

    @property
    def repo_url(self):
        return '{}/repos/{}/{}'.format(self.BASE_API_URL, self.owner, self.repo)

    def mock_edit_release(self, body=None, draft=True, prerelease=False):
        if body == None:
            body = 'Test release body'
        responses.add(
            method=responses.PATCH,
            url='{}/releases/1'.format(self.repo_url),
            json={
                'url': '{}/releases/1'.format(self.repo_url),
                'body': body,
                'draft': draft,
                'prerelease': prerelease,
            },
            status=httplib.OK,
        )

    def mock_get_repo(self):
        responses.add(
            method=responses.GET,
            url=self.repo_url,
            json={
                'url': self.repo_url,
            },
            status=httplib.OK,
        )

    def mock_list_pulls(self):
        responses.add(
            method=responses.GET,
            url='{}/pulls'.format(self.repo_url),
            json=[
                {
                    'id': 1,
                    'number': 1,
                },
            ],
            status=httplib.OK,
        )

    def mock_list_releases(self, tag=None, body=None):
        if tag == None:
            tag = 'v1.0'
        if body == None:
            body = 'Test release body'
        responses.add(
            method=responses.GET,
            url='{}/releases'.format(self.repo_url),
            json=[
                {
                    'tag_name': tag,
                    'url': '{}/releases/1'.format(self.repo_url),
                    'body': body,
                },
            ],
            status=httplib.OK,
        )

    def mock_post_comment(self, issue_number):
        responses.add(
            method=responses.POST,
            url='{}/issues/{}/comments'.format(self.repo_url, issue_number),
            status=httplib.OK,
        )

    def mock_pull_request(self, pr_number, body, title=None):
        if title == None:
            title = 'Test Pull Request Title'
        responses.add(
            method=responses.GET,
            url='{}/pulls/{}'.format(self.repo_url, pr_number),
            json={
                'base': {
                    'ref': 'master',
                },
                'body': body,
                'head': {
                    'ref': 'new-topic',
                },
                'html_url': '{}/pulls/{}'.format(self.html_url, pr_number),
                'issue_url': '{}/issues/{}'.format(self.repo_url, pr_number),
                'number': pr_number,
                'title': title,
            },
            status=httplib.OK,
        )



from __future__ import unicode_literals
from collections import OrderedDict
from distutils.version import LooseVersion
import os

from github3 import login
import hiyapyco
import raven

import cumulusci
from cumulusci.core.config import BaseTaskFlowConfig
from cumulusci.core.exceptions import ConfigError
from cumulusci.core.exceptions import DependencyResolutionError
from cumulusci.core.exceptions import KeychainNotFound
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import ServiceNotValid


class BaseProjectConfig(BaseTaskFlowConfig):
    """ Base class for a project's configuration which extends the global config """

    search_path = ['config']

    def __init__(self, global_config_obj, config=None):
        self.global_config_obj = global_config_obj
        self.keychain = None
        if not config:
            config = {}
        super(BaseProjectConfig, self).__init__(config=config)

    @property
    def config_global_local(self):
        return self.global_config_obj.config_global_local

    @property
    def config_global(self):
        return self.global_config_obj.config_global

    @property
    def repo_info(self):
        if hasattr(self, '_repo_info'):
            return self._repo_info

        # Detect if we are running in a CI environment and get repo info
        # from env vars for the enviornment instead of .git files
        info = {
            'ci': None
        }

        # Make sure that the CUMULUSCI_AUTO_DETECT environment variable is 
        # set before trying to auto-detect anything from the environment
        if not os.environ.get('CUMULUSCI_AUTO_DETECT'):
            self._repo_info = info
            return self._repo_info

        # Heroku CI
        heroku_ci = os.environ.get('HEROKU_TEST_RUN_ID')
        if heroku_ci:
            info = {
                'branch': os.environ.get('HEROKU_TEST_RUN_BRANCH'),
                'commit': os.environ.get('HEROKU_TEST_RUN_COMMIT_VERSION'),
                'ci': 'heroku',
                'root': '/app',
            }

        # Other CI environment implementations can be implemented here...

        # Apply CUMULUSCI_REPO_* environment variables last so they can
        # override and fill in missing values from the CI environment
        repo_branch = os.environ.get('CUMULUSCI_REPO_BRANCH')
        if repo_branch:
            if repo_branch != info.get('branch'):
                self.logger.info(
                    'CUMULUSCI_REPO_BRANCH found, using its value as the branch'
                )
            info['branch'] = repo_branch
        repo_commit = os.environ.get('CUMULUSCI_REPO_COMMIT')
        if repo_commit:
            if repo_commit != info.get('commit'):
                self.logger.info(
                    'CUMULUSCI_REPO_COMMIT found, using its value as the commit'
                )
            info['commit'] = repo_commit
        repo_root = os.environ.get('CUMULUSCI_REPO_ROOT')
        if repo_root:
            if repo_root != info.get('root'):
                self.logger.info(
                    'CUMULUSCI_REPO_ROOT found, using its value as the repo root'
                )
            info['root'] = repo_root
        repo_url = os.environ.get('CUMULUSCI_REPO_URL')
        if repo_url:
            if repo_url != info.get('url'):
                self.logger.info(
                    'CUMULUSCI_REPO_URL found, using its value as the repo url, owner, and name'
                )
            url_info = self._split_repo_url(repo_url)
            info.update(url_info)

        # If running in a CI environment, make sure we have all the needed
        # git info or throw a ConfigError
        if info['ci']:
            validate = OrderedDict((
                # <key>, <env var to manually override>
                ('branch', 'CUMULUSCI_REPO_BRANCH'),
                ('commit', 'CUMULUSCI_REPO_COMMIT'),
                ('name', 'CUMULUSCI_REPO_URL'),
                ('owner', 'CUMULUSCI_REPO_URL'),
                ('root', 'CUMULUSCI_REPO_ROOT'),
                ('url', 'CUMULUSCI_REPO_URL'),
            ))
            for key, env_var in list(validate.items()):
                if key not in info or not info[key]:
                    message = 'Detected CI on {} but could not determine the repo {}'.format(
                        info['ci'],
                        key,
                    )
                    if env_var:
                        message += '. You can manually pass in the {} with'.format(key)
                        message += ' with the {} environment variable.'.format(env_var)
                    raise ConfigError(message)

        # Log any overrides detected through the environment as a warning
        if len(info) > 1:
            self.logger.info('')
            self.logger.warn(
                'Using environment variables to override repo info:'
            )
            keys = list(info.keys())
            keys.sort()
            for key in keys:
                self.logger.warn(
                    '  {}: {}'.format(key, info[key])
                )
            self.logger.info('')

        self._repo_info = info
        return self._repo_info

    def _split_repo_url(self, url):
        url_parts = url.split('/')
        name = url_parts[-1]
        owner = url_parts[-2]
        if name.endswith('.git'):
            name = name[:-4]
        git_info = {
            'url': url,
            'owner': owner,
            'name': name,
        }
        return git_info

    @property
    def repo_root(self):
        path = self.repo_info.get('root')
        if path:
            return path

        path = os.path.splitdrive(os.getcwd())[1]
        while True:
            if os.path.isdir(os.path.join(path, '.git')):
                return path
            head, tail = os.path.split(path)
            if not tail:
                # reached the root
                break
            path = head

    @property
    def repo_name(self):
        name = self.repo_info.get('name')
        if name:
            return name

        if not self.repo_root:
            return

        in_remote_origin = False
        with open(os.path.join(self.repo_root, '.git', 'config'), 'r') as f:
            for line in f:
                line = line.strip()
                if line == '[remote "origin"]':
                    in_remote_origin = True
                    continue
                if in_remote_origin and line.find('url =') != -1:
                    return self._split_repo_url(line)['name']

    @property
    def repo_url(self):
        url = self.repo_info.get('url')
        if url:
            return url

        if not self.repo_root:
            return

        git_config_file = os.path.join(self.repo_root, '.git', 'config')
        with open(git_config_file, 'r') as f:
            in_remote_origin = False
            for line in f:
                line = line.strip()
                if line == '[remote "origin"]':
                    in_remote_origin = True
                    continue
                if in_remote_origin and 'url = ' in line:
                    return line[7:]

    @property
    def repo_owner(self):
        owner = self.repo_info.get('owner')
        if owner:
            return owner

        if not self.repo_root:
            return

        in_remote_origin = False
        with open(os.path.join(self.repo_root, '.git', 'config'), 'r') as f:
            for line in f:
                line = line.strip()
                if line == '[remote "origin"]':
                    in_remote_origin = True
                    continue
                if in_remote_origin and line.find('url =') != -1:
                    line_parts = line.split('/')
                    return line_parts[-2].split(':')[-1]

    @property
    def repo_branch(self):
        branch = self.repo_info.get('branch')
        if branch:
            return branch

        if not self.repo_root:
            return

        with open(os.path.join(self.repo_root, '.git', 'HEAD'), 'r') as f:
            branch_ref = f.read().strip()
        if branch_ref.startswith('ref: '):
            return '/'.join(branch_ref[5:].split('/')[2:])

    @property
    def repo_commit(self):
        commit = self.repo_info.get('commit')
        if commit:
            return commit

        if not self.repo_root:
            return

        branch = self.repo_branch
        if not branch:
            return

        join_args = [self.repo_root, '.git', 'refs', 'heads']
        join_args.extend(branch.split('/'))
        commit_file = os.path.join(*join_args)

        commit_sha = None
        if os.path.isfile(commit_file):
            with open(commit_file, 'r') as f:
                commit_sha = f.read().strip()
        else:
            packed_refs_path = os.path.join(
                self.repo_root,
                '.git',
                'packed-refs'
            )
            with open(packed_refs_path, 'r') as f:
                for line in f:
                    parts = line.split(' ')
                    if len(parts) == 1:
                        # Skip lines showing the commit sha of a tag on the
                        # preceeding line
                        continue
                    if parts[1].replace('refs/remotes/origin/', '').strip() == branch:
                        commit_sha = parts[0]
                        break

        return commit_sha

    @property
    def use_sentry(self):
        try:
            self.keychain.get_service('sentry')
            return True
        except ServiceNotConfigured:
            return False
        except ServiceNotValid:
            return False

    def init_sentry(self, ):
        """ Initializes sentry.io error logging for this session """
        if not self.use_sentry:
            return

        sentry_config = self.keychain.get_service('sentry')

        tags = {
            'repo': self.repo_name,
            'branch': self.repo_branch,
            'commit': self.repo_commit,
            'cci version': cumulusci.__version__,
        }
        tags.update(self.config.get('sentry_tags', {}))

        env = self.config.get('sentry_environment', 'CumulusCI CLI')

        self.sentry = raven.Client(
            dsn=sentry_config.dsn,
            environment=env,
            tags=tags,
            processors=(
                'raven.processors.SanitizePasswordsProcessor',
            ),
        )

    def get_github_api(self):
        github_config = self.keychain.get_service('github')
        gh = login(github_config.username, github_config.password)
        return gh

    def get_latest_version(self, beta=False):
        """ Query Github Releases to find the latest production or beta release """
        gh = self.get_github_api()
        repo = gh.repository(self.repo_owner, self.repo_name)
        latest_version = None
        for release in repo.iter_releases():
            if beta != release.tag_name.startswith(self.project__git__prefix_beta):
                continue
            version = self.get_version_for_tag(release.tag_name)
            if version is None:
                continue
            version = LooseVersion(version)
            if not latest_version or version > latest_version:
                latest_version = version
        return latest_version

    @property
    def config_project_path(self):
        if not self.repo_root:
            return
        path = os.path.join(self.repo_root, self.config_filename)
        if os.path.isfile(path):
            return path

    @property
    def project_local_dir(self):
        """ location of the user local directory for the project
        e.g., ~/.cumulusci/NPSP-Extension-Test/ """

        # depending on where we are in bootstrapping the YamlGlobalConfig
        # the canonical projectname could be located in one of two places
        if self.project__name:
            name = self.project__name
        else:
            try:
                name = self.config_project['project']['name']
            except KeyError:
                name = ''

        path = os.path.join(
            os.path.expanduser('~'),
            self.global_config_obj.config_local_dir,
            name,
        )
        if not os.path.isdir(path):
            os.makedirs(path)
        return path

    def get_tag_for_version(self, version):
        if '(Beta' in version:
            tag_version = version.replace(
                ' (', '-').replace(')', '').replace(' ', '_')
            tag_name = self.project__git__prefix_beta + tag_version
        else:
            tag_name = self.project__git__prefix_release + version
        return tag_name

    def get_version_for_tag(self, tag, prefix_beta=None, prefix_release=None):
        if prefix_beta is None:
            prefix_beta = self.project__git__prefix_beta
        if prefix_release is None:
            prefix_release = self.project__git__prefix_release
        if tag.startswith(prefix_beta):
            version = tag.replace(prefix_beta, '')
            if '-Beta_' in version:
                # Beta tags are expected to be like "beta/1.0-Beta_1"
                # which is returned as "1.0 (Beta 1)"
                return version.replace('-', ' (').replace('_', ' ') + ')'
            else:
                return
        elif tag.startswith(prefix_release):
            return tag.replace(prefix_release, '')

    def set_keychain(self, keychain):
        self.keychain = keychain

    def _check_keychain(self):
        if not self.keychain:
            raise KeychainNotFound(
                'Could not find config.keychain. You must call ' +
                'config.set_keychain(keychain) before accessing orgs')

    def list_orgs(self):
        """ Returns a list of all org names for the project """
        self._check_keychain()
        return self.keychain.list_orgs()

    def get_org(self, name):
        """ Returns an OrgConfig for the given org_name """
        self._check_keychain()
        return self.keychain.get_org(name)

    def set_org(self, name, org_config):
        """ Creates or updates an org's oauth info """
        self._check_keychain()
        return self.keychain.set_org(name, org_config)

    def get_static_dependencies(self, dependencies=None):
        """ Resolves the project -> dependencies section of cumulusci.yml
            to convert dynamic github dependencies into static dependencies
            by inspecting the referenced repositories
        """
        if not dependencies:
            dependencies = self.project__dependencies

        if not dependencies:
            return

        static_dependencies = []
        for dependency in dependencies:
            if 'github' not in dependency:
                static_dependencies.append(dependency)
            else:
                static = self.process_github_dependency(dependency)
                static_dependencies.extend(static)
        return static_dependencies

    def pretty_dependencies(self, dependencies, indent=None):
        if not indent:
            indent = 0
        pretty = []
        for dependency in dependencies:
            prefix = '{}  - '.format(" " * indent)
            for key, value in list(dependency.items()):
                extra = []
                if value is None or value is False:
                    continue
                if key == 'dependencies':
                    extra = self.pretty_dependencies(
                        dependency['dependencies'], indent=indent + 4)
                    if not extra:
                        continue
                    value = '\n{}'.format(" " * (indent + 4))

                pretty.append('{}{}: {}'.format(prefix, key, value))
                if extra:
                    pretty.extend(extra)
                prefix = '{}    '.format(" " * indent)
        return pretty

    def process_github_dependency(self, dependency, indent=None):
        if not indent:
            indent = ''

        self.logger.info(
            '{}Processing dependencies from Github repo {}'.format(
                indent,
                dependency['github'],
            )
        )

        skip = dependency.get('skip')
        if not isinstance(skip, list):
            skip = [skip, ]

        # Initialize github3.py API against repo
        gh = self.get_github_api()
        repo_owner, repo_name = dependency['github'].split('/')[3:5]
        if repo_name.endswith('.git'):
            repo_name = repo_name[:-4]
        repo = gh.repository(repo_owner, repo_name)

        # Prepare HTTP auth header for requests calls to Github
        github = self.keychain.get_service('github')
        headers = {'Authorization': 'token {}'.format(github.password)}

        # Determine the ref if specified
        kwargs = {}
        if 'tag' in dependency:
            tag = dependency['tag']
            kwargs['ref'] = tag
        else:
            tag = None

        # Get the cumulusci.yml file
        contents = repo.contents('cumulusci.yml', **kwargs)
        cumulusci_yml = hiyapyco.load(contents.decoded, loglevel='INFO')

        # Get the namespace from the cumulusci.yml if set
        namespace = cumulusci_yml.get('project', {}).get(
            'package', {}).get('namespace')

        # Check for unmanaged flag on a namespaced package
        unmanaged = namespace and dependency.get('unmanaged') is True

        # Look for subfolders under unpackaged/pre
        unpackaged_pre = []
        contents = repo.contents('unpackaged/pre', **kwargs)
        if contents:
            for dirname in list(contents.keys()):
                if 'unpackaged/pre/{}'.format(dirname) in skip:
                    continue
                subfolder = "{}-{}/unpackaged/pre/{}".format(
                    repo.name, repo.default_branch, dirname)
                zip_url = "{}/archive/{}.zip".format(
                    repo.html_url, repo.default_branch)

                unpackaged_pre.append({
                    'zip_url': zip_url,
                    'subfolder': subfolder,
                    'headers': headers,
                    'unmanaged': dependency.get('unmanaged'),
                    'namespace_tokenize': dependency.get('namespace_tokenize'),
                    'namespace_inject': dependency.get('namespace_inject'),
                    'namespace_strip': dependency.get('namespace_strip'),
                })

        # Look for metadata under src (deployed if no namespace)
        unmanaged_src = None
        if unmanaged or not namespace:
            contents = repo.contents('src', **kwargs)
            if contents:
                zip_url = "{}/archive/{}.zip".format(
                    repo.html_url, repo.default_branch)
                subfolder = "{}-{}/src".format(repo.name, repo.default_branch)

                unmanaged_src = {
                    'zip_url': zip_url,
                    'subfolder': subfolder,
                    'headers': headers,
                    'unmanaged': dependency.get('unmanaged'),
                    'namespace_tokenize': dependency.get('namespace_tokenize'),
                    'namespace_inject': dependency.get('namespace_inject'),
                    'namespace_strip': dependency.get('namespace_strip'),
                }

        # Look for subfolders under unpackaged/post
        unpackaged_post = []
        contents = repo.contents('unpackaged/post', **kwargs)
        if contents:
            for dirname in list(contents.keys()):
                if 'unpackaged/post/{}'.format(dirname) in skip:
                    continue
                zip_url = "{}/archive/{}.zip".format(
                    repo.html_url, repo.default_branch)
                subfolder = "{}-{}/unpackaged/post/{}".format(
                    repo.name, repo.default_branch, dirname)

                dependency = {
                    'zip_url': zip_url,
                    'subfolder': subfolder,
                    'headers': headers,
                    'unmanaged': dependency.get('unmanaged'),
                    'namespace_tokenize': dependency.get('namespace_tokenize'),
                    'namespace_inject': dependency.get('namespace_inject'),
                    'namespace_strip': dependency.get('namespace_strip'),
                }
                # By default, we always inject the project's namespace into
                # unpackaged/post metadata
                if namespace and not dependency.get('namespace_inject'):
                    dependency['namespace_inject'] = namespace
                    dependency['unmananged'] = unmanaged
                unpackaged_post.append(dependency)

        # Parse values from the repo's cumulusci.yml
        project = cumulusci_yml.get('project', {})
        prefix_beta = project.get('git', {}).get('prefix_beta', 'beta/')
        prefix_release = project.get('git', {}).get('prefix_release', 'release/')
        dependencies = project.get('dependencies')
        if dependencies:
            dependencies = self.get_static_dependencies(dependencies)

        # Create the final ordered list of all parsed dependencies
        repo_dependencies = []

        # unpackaged/pre/*
        if unpackaged_pre:
            repo_dependencies.extend(unpackaged_pre)

        if namespace and not unmanaged:
            version = None
            if tag:
                version = self.get_version_for_tag(tag, prefix_beta, prefix_release)
            else:
                # github3.py doesn't support the latest release API so we hack
                # it together here
                url = repo._build_url('releases/latest', base_url=repo._api)
                try:
                    version = repo._get(url).json()['name']
                except Exception as e:
                    self.logger.warn('{}{}: {}'.format(
                        indent, e.__class__.__name__, e.message))

            if not version:
                raise DependencyResolutionError(
                    '{}Could not find latest release for {}'.format(indent, namespace)
                )
            # If a latest prod version was found, make the dependencies a
            # child of that install
            dependency = {
                'namespace': namespace,
                'version': version,
            }
            if dependencies:
                dependency['dependencies'] = dependencies
            repo_dependencies.append(dependency)

        # Unmanaged metadata from src (if referenced repo doesn't have a
        # namespace)
        else:
            if dependencies:
                repo_dependencies.extend(dependencies)
            if unmanaged_src:
                repo_dependencies.append(unmanaged_src)

        # unpackaged/post/*
        if unpackaged_post:
            repo_dependencies.extend(unpackaged_post)

        return repo_dependencies



from cumulusci.tasks.salesforce import BaseSalesforceTask


class BaseSalesforceMetadataApiTask(BaseSalesforceTask):
    api_class = None
    name = 'BaseSalesforceMetadataApiTask'

    def _get_api(self):
        return self.api_class(self)

    def _run_task(self):
        api = self._get_api()
        if api:
            return api()
